DEFAULT_SIGNED_URL_CREATOR = "https://europe-west2-chai-959f8.cloudfunctions.net/signed-url-creator"
GUEST_UID = "20c900f0-e6f5-440c-bd83-e1ad894ad598"
GUEST_KEY = "LyRgaRjLmldLltvIL5Q5QY6XxNgIAMiTxxQdq2GyiJA8jsd55q7FxYcOgZUbV2dFmeK75pt4vd3zXe9Acd7pyw"

BOT_DELETER_ENDPOINT = "https://europe-west2-chai-959f8.cloudfunctions.net/bot-deleter"
DEFAULT_BOT_STATUS_ENDPOINT = "https://europe-west2-chai-959f8.cloudfunctions.net/bot-status"
DEFAULT_BOT_LOGS_ENDPOINT = "https://europe-west2-chai-959f8.cloudfunctions.net/bot-logs"
ENDPOINT = "https://europe-west2-chai-959f8.cloudfunctions.net/chai-api"

API_HOST = 'http://api.chai.ml'
